import pandas as pd
import geopandas as gpd
import numpy as np
from .graph import Graph
from ..util import transform
import logging
from math import ceil, floor, sqrt


class BusSim:

    def __init__(
        self,
        manager,
        day,
        start_time,
        elapse_time,
        avg_walking_speed=1.4,
        max_walking_min=-1,  # HACK
        trip_delays=[]
    ):
        """The constructor of the BusSim class

        Args:

            day (str): the day in a week to perform simulation on

            start_time (str): the starting time to perform simulation on
                (HH:MM:SS)

            elapse_time (str): the elapse time from starting time to perform
                simulation on (HH:MM:SS)

            avg_walking_speed (float): the assumed average walking speed

            max_walking_min (float): the maximum allowed walking time minutes

            trip_delays (list[Tuple], optional): the list
                of trip-delay pairs to add the tuple should be in the format of
                `(trip_id, delay in HH:MM:SS)`

        """
        self._logger = logging.getLogger('app')
        self._logger.info("Start initializing sim")
        self.manager = manager
        self.day = day
        self.start_time = start_time
        self.elapse_time = elapse_time
        self.avg_walking_speed = avg_walking_speed

        # HACK
        if max_walking_min == -1:
            max_walking_min = elapse_time
        self.max_walking_min = max_walking_min
        self.max_walking_distance = max_walking_min * 60.0 * avg_walking_speed
        self.stopTimes_final_df = self._gen_final_df(trip_delays)
        self.graph = Graph(self.stopTimes_final_df, start_time,
                           elapse_time, self.max_walking_distance, avg_walking_speed)
        self._logger.info("Sim successfully initialized")

    def get_access_grid(self, start_stop=None, start_point=None, grid_size_min=2, route_remove=[]):
        max_x, min_x, max_y, min_y = self.manager.get_borders()
        x_num, y_num, grid_size = self._get_grid_dimention(grid_size_min)
        grid = np.zeros(x_num*y_num).reshape(y_num, -1)

        self._logger.info("Start searching graph")
        # first convert start_point into meters
        if start_point is not None:
            start_point = transform(start_point[0], start_point[1])
        stops_radius_list = self.graph.search(
            start_stop, start_point, route_remove)

        if stops_radius_list is None or len(stops_radius_list) == 0:
            return grid

        self._logger.info("Start compressing")
        for bubble in stops_radius_list:
            min_x_idx = floor(
                (bubble["stop_x"] - min_x - bubble["radius"]) / grid_size)
            max_x_idx = floor(
                (bubble["stop_x"] - min_x + bubble["radius"]) / grid_size)
            min_y_idx = floor(
                (bubble["stop_y"] - min_y - bubble["radius"]) / grid_size)
            max_y_idx = floor(
                (bubble["stop_y"] - min_y + bubble["radius"]) / grid_size)

            # TODO: make this radius calculation more precise
            radius_idx = (max_x_idx - min_x_idx + 1) / 2
            mid_x_idx = (max_x_idx + min_x_idx) / 2
            mid_y_idx = (max_y_idx + min_y_idx) / 2

            start_y = max(0, min_y_idx)
            start_x = max(0, min_x_idx)
            end_y = min(max_y_idx+1, y_num)
            end_x = min(max_x_idx+1, x_num)

            for y in range(start_y, end_y):
                for x in range(start_x, end_x):
                    # check 4 corners, if the distance for all of them is greater than radius, then this grid is not in the circle
                    if (y - mid_y_idx) ** 2 + (x - mid_x_idx) ** 2 < radius_idx ** 2 or \
                        (y - mid_y_idx) ** 2 + (x + 1 - mid_x_idx) ** 2 < radius_idx ** 2 or \
                        (y + 1 - mid_y_idx) ** 2 + (x - mid_x_idx) ** 2 < radius_idx ** 2 or \
                            (y + 1 - mid_y_idx) ** 2 + (x + 1 - mid_x_idx) ** 2 < radius_idx ** 2:
                        grid[y][x] = 1

        self._logger.info("Finish generating grid")
        return grid

    def get_gdf(self, start_stop=None, start_point=None, route_remove=[]):
        """Given a starting point(lat, lon) or a starting stop_id, compute the region covered in geopandas.Geodataframe

        Args:
            start_stop (str, optional): The path to the directory of the data files
                (contains both mmt_gtfs and plot subdirectories)

            start_point (str, optional): the day in a week to perform simulation on

        Returns:
            geopandas.GeoDataFrame: the GeoDataFrame of the region covered

        """
        self._logger.info("Start searching graph")

        # first convert start_point into meters
        if start_point is not None:
            start_point = transform(start_point[0], start_point[1])
        stops_radius_list = self.graph.search(
            start_stop, start_point, route_remove)

        if stops_radius_list is None or len(stops_radius_list) == 0:
            return

        self._logger.debug("start generating gdf")
        df = pd.DataFrame(stops_radius_list)

        gdf = gpd.GeoDataFrame(
            df, geometry=gpd.points_from_xy(df.stop_x, df.stop_y), crs="EPSG:3174")

        self._logger.debug("start generating geometry buffer with radius")
        gdf['geometry'] = gdf.geometry.buffer(gdf['radius'])
        self._logger.info("Finish generating gdf")
        return gdf

    # def get_area(self, gdf):
    #     """This is a util method to compute the total area in meters^2 given a geopandas.Geodataframe

    #     Args:
    #         gdf (geopandas.GeoDataFrame): The Geodataframe used to compute the total area

    #     Returns:
    #         float: the total area in meters^2

    #     """
    #     if gdf is None:
    #         return

    #     # the area returned is in meters^2
    #     self._logger.info("start calculating area")

    #     self._logger.debug("start calculating union/difference")
    #     area = gdf.unary_union.difference(self.lakes.unary_union).area
    #     self._logger.info("finish calculating area")
    #     return area

    def _gen_final_df(self, trip_delays):
        self._logger.debug("Start generating dataframe")

        stops_df = self.manager.read_gtfs("stops-3174.txt")
        trips_df = self.manager.read_gtfs("trips.txt")
        stopTimes_df = self.manager.read_gtfs("stop_times.txt")
        calendar_df = self.manager.read_gtfs("calendar.txt")

        # get valid service_ids
        calendar_df['start_date'] = pd.to_datetime(
            calendar_df['start_date'], format='%Y%m%d')
        calendar_df['end_date'] = pd.to_datetime(
            calendar_df['end_date'], format='%Y%m%d')
        calendar_filtered_df = calendar_df[self._is_service_valid(
            calendar_df[self.day], calendar_df["service_id"])]
        service_ids = calendar_filtered_df["service_id"].tolist()

        # get valid trips
        trips_df = trips_df[trips_df["service_id"].isin(service_ids)]

        # get valid stop_times
        stopTimes_filtered_df = trips_df.merge(
            stopTimes_df, on="trip_id")
        stopTimes_merged_df = stopTimes_filtered_df.merge(stops_df, on="stop_id")[
            ["service_id", "route_short_name", "trip_id", "stop_id", "stop_sequence", "arrival_time", "shape_dist_traveled", "stop_x", "stop_y", "cardinal_direction"]]

        # get stop_times within the time frame
        stopTimes_merged_df['arrival_time'] = pd.to_timedelta(
            stopTimes_merged_df['arrival_time'])

        # add trip_delays
        for (trip_id, delay) in trip_delays:
            stopTimes_merged_df.loc[stopTimes_merged_df["trip_id"]
                                    == trip_id, "arrival_time"] += pd.to_timedelta(delay)

        stopTimes_final_df = self._get_valid_stopTime(
            stopTimes_merged_df, self.start_time, self.elapse_time).sort_values(by="arrival_time")

        return stopTimes_final_df

    def get_available_route(self):
        return self.stopTimes_final_df["route_short_name"].unique()

    def _is_service_valid(self, day, service_id):
        # FIXME: hardcode in the service to be 95, just pick the first service id
        return (day == 1) & (service_id.str.startswith("95"))

    def _get_valid_stopTime(self, df, start_time, elapse_time):
        start_time = pd.to_timedelta(start_time)
        end_time = start_time + pd.to_timedelta(elapse_time)
        return df[(df['arrival_time'] > start_time) & (df['arrival_time'] < end_time)]

    def _get_grid_dimention(self, grid_size_min):
        max_x, min_x, max_y, min_y = self.manager.get_borders()
        grid_size = grid_size_min * self.avg_walking_speed * 60
        x_num = ceil(abs(max_x - min_x) / grid_size)
        y_num = ceil(abs(max_y - min_y) / grid_size)
        return x_num, y_num, grid_size



from datetime import timedelta
from math import sin, cos, asin, sqrt, pi, ceil, floor
from collections import defaultdict
import heapq
import pandas as pd
import geopandas as gpd
import logging


class Node:
    def __init__(self, trip_id, route_short_name, stop_sequence, stop_id, stop_x, stop_y, arrival_time, max_walking_distance):
        self.trip_id = trip_id
        self.route_short_name = route_short_name
        self.stop_sequence = stop_sequence
        self.stop_id = stop_id
        self.stop_x = stop_x
        self.stop_y = stop_y
        self.arrival_time = arrival_time
        # this should be modified by search in graph
        self.walking_distance = max_walking_distance
        self.children = []

    def distance(self, other):
        return sqrt((other.stop_x - self.stop_x)**2 + (other.stop_y - self.stop_y)**2)

    # Deprecated
    def harversine_distance(self, other):
        """Calculates the distance between two points on earth using the
        harversine distance (distance between points on a sphere)
        See: https://en.wikipedia.org/wiki/Haversine_formula

        :return: distance in meters between points
        """

        lat1, lon1, lat2, lon2 = (
            a/180*pi for a in [self.stop_lat, self.stop_lon, other.stop_lat, other.stop_lon])
        dlon = lon2 - lon1
        dlat = lat2 - lat1
        a = sin(dlat/2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon/2) ** 2
        c = 2 * asin(min(1, sqrt(a)))
        d = 3956 * 1609.344 * c
        return d

    def __str__(self):
        return f"({self.trip_id}, {self.route_short_name}, {self.stop_sequence}, {self.stop_id}, {self.arrival_time}, {self.walking_distance}"

    def __repr__(self):
        rv = self.__str__()
        rv += "\nChildren:\n"
        for child in self.children:
            rv += f"  cost: {child.cost} "
            child = child.node
            rv += str(child)
            rv += "\n"
        return rv

    def __lt__(self, other):
        # always retain sequence here
        return False


class NodeCostPair:
    def __init__(self, node, cost):
        self.node = node
        self.cost = cost  # the cost here means the walking distance


class Graph:
    def __init__(self, df, start_time, elapse_time, max_walking_distance, avg_walking_speed):
        self._logger = logging.getLogger('app')
        self._logger.debug("start generating graph")

        self.df = df
        self.start_time = start_time
        self.elapse_time = elapse_time
        self.max_walking_distance = max_walking_distance
        self.avg_walking_speed = avg_walking_speed
        self.nodes = []
        self.empty = False
        self._constuct_graph()

        self._logger.debug(f"generated {len(self.nodes)} nodes in the graph")

    # return the area coverage data after performing graph search
    # return format: [{“stop_id”, “stop_y”, “stop_x”, “radius”}, ... ]
    def search(self, start_stop=None, start_point=None, route_remove=[]):
        if self.empty:
            return

        self._logger.debug("start locating starting stop")
        start = self._find_start(start_stop, start_point)
        self._logger.debug("start clearing graph")
        self._clear_graph()
        self._logger.debug("start runnning dijkstra")
        self._dijkstra(start, route_remove)

        self._logger.debug("start collecting nodes")
        stops_radius_dict = dict()
        start_time = pd.to_timedelta(self.start_time)
        end_time = start_time + pd.to_timedelta(self.elapse_time)
        for node in self.nodes:
            if node.walking_distance < self.max_walking_distance:
                radius = self.max_walking_distance - node.walking_distance
                time_left = (end_time - node.arrival_time).total_seconds()
                radius = min(radius, self.avg_walking_speed * time_left)
                if node.stop_id not in stops_radius_dict or radius > stops_radius_dict[node.stop_id]["radius"]:
                    stops_radius_dict[node.stop_id] = {
                        "stop_id": node.stop_id,
                        "stop_x": node.stop_x,
                        "stop_y": node.stop_y,
                        "radius": radius
                    }

        stops_radius_list = [row for row in stops_radius_dict.values()]
        return stops_radius_list

    def _clear_graph(self):
        for node in self.nodes:
            node.walking_distance = self.max_walking_distance

    def _dijkstra(self, start, route_remove):
        pq = [(0, start)]
        while len(pq) > 0:
            curr_distance, curr_node = heapq.heappop(pq)
            # print(f"curr_distance: {curr_distance} curr_node: {curr_node}")

            if curr_distance > curr_node.walking_distance:
                continue

            for child in curr_node.children:
                if child.node.route_short_name in route_remove:
                    continue
                cost = child.cost
                child = child.node

                distance = curr_distance + cost

                if distance < child.walking_distance:
                    child.walking_distance = distance
                    heapq.heappush(pq, (distance, child))

    def _constuct_graph(self):
        if len(self.df) == 0:
            self.empty = True
            return

        # gen nodes
        trip_node_dict = defaultdict(list)
        stop_node_dict = defaultdict(list)

        map_grid = []
        min_x = self.df.stop_x.min()
        max_x = self.df.stop_x.max()
        min_y = self.df.stop_y.min()
        max_y = self.df.stop_y.max()

        x_num = ceil((max_x - min_x) / self.max_walking_distance)
        y_num = ceil((max_y - min_y) / self.max_walking_distance)
        for i in range(x_num):
            x_list = []
            for j in range(y_num):
                x_list.append([])
            map_grid.append(x_list)

        for index, row in self.df.iterrows():
            node = Node(row["trip_id"], row["route_short_name"], row["stop_sequence"], row["stop_id"], row["stop_x"],
                        row["stop_y"], row["arrival_time"], self.max_walking_distance)
            self.nodes.append(node)
            trip_node_dict[row["trip_id"]].append(node)
            stop_node_dict[row["stop_id"]].append(node)
            x_bucket = floor((row["stop_x"] - min_x) /
                             self.max_walking_distance)
            y_bucket = floor((row["stop_y"] - min_y) /
                             self.max_walking_distance)
            map_grid[x_bucket][y_bucket].append(node)

        # gen edges
        # direct sequence
        for trip_id, nodes in trip_node_dict.items():
            for i in range(len(nodes)-1):
                start = nodes[i]
                end = nodes[i+1]
                nodeCostPair = NodeCostPair(end, 0)
                start.children.append(nodeCostPair)

        # wait on the stop
        for stop_id, nodes in stop_node_dict.items():
            for i in range(len(nodes)):
                for j in range(len(nodes)):
                    start = nodes[i]
                    end = nodes[j]
                    if start.arrival_time < end.arrival_time:
                        nodeCostPair = NodeCostPair(end, 0)
                        start.children.append(nodeCostPair)

        # walk
        for x in range(x_num):
            for y in range(y_num):
                start_bucket = map_grid[x][y]
                end_buckets = []
                for x_end in range(max(0, x-1), min(x_num, x+2)):
                    for y_end in range(max(0, y-1), min(y_num, y+2)):
                        end_buckets.append(map_grid[x_end][y_end])

                for start in start_bucket:
                    for end_bucket in end_buckets:
                        for end in end_bucket:
                            if start.arrival_time >= end.arrival_time or start.stop_id == end.stop_id:
                                continue

                            # walk
                            distance = start.distance(end)
                            time_delta = distance / self.avg_walking_speed
                            time_delta = timedelta(seconds=time_delta)
                            if distance < self.max_walking_distance and start.arrival_time + time_delta < end.arrival_time:
                                nodeCostPair = NodeCostPair(end, distance)
                                start.children.append(nodeCostPair)

    def _find_start(self, start_stop, start_point):
        if start_stop is not None:
            return self._find_start_stop(start_stop)

        elif start_point is not None:
            return self._find_start_point(start_point)

    def _find_start_stop(self, start_stop):
        for node in self.nodes:
            if node.stop_id == start_stop:
                start_point = (node.stop_x, node.stop_y)
                return self._find_start_point(start_point)

    def _find_start_point(self, start_point):
        x, y = start_point
        start = Node(None, None, None, None, x, y,
                     pd.to_timedelta(self.start_time), 0)

        # gen edges by walking
        for end in self.nodes:
            # unreachable for sure (can't go back in time)
            if start.arrival_time >= end.arrival_time:
                continue

            # walk
            distance = start.distance(end)
            time_delta = distance / self.avg_walking_speed
            time_delta = timedelta(seconds=time_delta)
            if distance < self.max_walking_distance and start.arrival_time + time_delta < end.arrival_time:
                nodeCostPair = NodeCostPair(end, distance)
                start.children.append(nodeCostPair)

        return start



from .managerFactory import *
from .localManager import *
from .AWSManager import *

managerFactory = ManagerFactory()
managerFactory.register_manager("local", LocalManager)
managerFactory.register_manager("aws", AWSManager)



from .manager import managerFactory
from ..util import tomin
from shapely.geometry import Point


class Config:
    def __init__(self, day, elapse_time, interval, avg_walking_speed=1.4, max_walking_min=None, grid_size_min=2, route_remove=[], run_env="local"):
        self.params_ = {
            "run_env": run_env,
            "busSim_params": {
                "day": day,
                "elapse_time": elapse_time,
                "avg_walking_speed": avg_walking_speed,
                "max_walking_min": max_walking_min if max_walking_min != None else tomin(elapse_time),
                "grid_size_min": grid_size_min,
                "route_remove": route_remove
            },
            "interval": interval,
            "start_points": []
        }

    def set_starts(self, points=[], centroids=None):
        if centroids is not None:
            self._set_starts_from_centroids(centroids)
        if len(points) > 0:
            self._set_starts_from_points(points)

    def get_start_points(self):
        return self.params_.get("start_points")

    def get_run_env(self):
        return self.params_.get("run_env")

    def get_interval(self):
        return self.params_.get("interval")

    def get_busSim_params(self):
        return self.params_.get("busSim_params")

    def is_runnable(self):
        # check run_env
        if self.params_.get("run_env") not in managerFactory.get_available_managers():
            return False

        # check start_points
        if len(self.params_.get("start_points")) == 0:
            return False

        return True

    def _set_starts_from_points(self, points):
        for point in points:
            if type(point) is tuple:
                self.params_["start_points"].append(point)
            elif type(point) is Point:
                self.params_["start_points"].append((point.y, point.x))

    def _set_starts_from_centroids(self, centroids):
        for poly in centroids["geometry"]:
            point = poly.centroid
            self.params_["start_points"].append((point.y, point.x))



import os
from ..util import tomin


class SearchResult():
    BITMASK = [0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80]
    DAY = ["monday", "tuesday", "wednesday",
           "thursday", "friday", "saturday", "sunday"]

    def __init__(self, busSim, grid_size_min):
        self.data = b""
        self.length = 0
        self.day = self.DAY.index(busSim.day)
        self.start_time = tomin(busSim.start_time)
        self.elapse_time = tomin(busSim.elapse_time)
        # self.avg_walking_speed = busSim.avg_walking_speed
        self.max_walking_min = busSim.max_walking_min
        self.grid_size_min = grid_size_min
        self.x_num, self.y_num, _ = busSim._get_grid_dimention(grid_size_min)
        self.data_block_size = (self.x_num * self.y_num + 7) // 8

    @classmethod
    def load_meta(cls, f):
        meta = {}
        meta["day"] = f.read(1)
        meta["start_time"] = f.read(2)
        meta["elapse_time"] = f.read(1)
        meta["max_walking_min"] = f.read(1)
        meta["grid_size_min"] = int.from_bytes(f.read(1), "big")
        meta["x_num"] = int.from_bytes(f.read(2), "big")
        meta["y_num"] = int.from_bytes(f.read(2), "big")
        meta["length"] = int.from_bytes(f.read(2), "big")
        return meta

    @classmethod
    def load_bitmap(cls, f, x_num, y_num, data_block_size):
        grid = []
        row = []
        data = f.read(data_block_size)

        for i in range(x_num * y_num):
            x = i % x_num
            if x == 0 and i != 0:
                grid.append(row)
                row = []

            row.append(data[i // 8] >> (i % 8) & 0x01)

        grid.append(row)
        return grid

    @classmethod
    def load_grid(cls, filename, idx):
        with open(filename, "rb") as f:
            meta = cls.load_meta(f)
            # TODO: fix this hardcoded 1.4 avg walking speed. could save this in the metadata
            grid_size = meta["grid_size_min"]*1.4*60
            data_block_size = (meta["x_num"] * meta["y_num"] + 7) // 8
            f.seek(data_block_size*idx, 1)  # seek to the target bitmap
            grid = cls.load_bitmap(
                f, meta["x_num"], meta["y_num"], data_block_size)

        return grid, grid_size

    @classmethod
    def grid_iter(cls, filename):
        with open(filename, "rb") as f:
            meta = cls.load_meta(f)
            # TODO: fix this hardcoded 1.4 avg walking speed. could save this in the metadata
            grid_size = meta["grid_size_min"]*1.4*60
            data_block_size = (meta["x_num"] * meta["y_num"] + 7) // 8
            for i in range(meta["length"]):
                yield cls.load_bitmap(f, meta["x_num"], meta["y_num"], data_block_size), grid_size

    def get_out_filename(self):
        return f"search-result-{self.day}-{self.start_time}"

    def record(self, start_point, grid):
        data = self._serialize_grid(grid)
        self.length += 1
        self.data += data

    def to_bytes(self):
        header = self._serialize_header()
        return header + self.data

    def _serialize_grid(self, grid):
        data = bytearray(self.data_block_size)
        for y, row in enumerate(grid):
            for x, bit in enumerate(row):
                if bit == 1:
                    pos = y * self.x_num + x
                    data[pos // 8] |= self.BITMASK[pos % 8]
        return data

    def _serialize_header(self):
        return self.day.to_bytes(1, byteorder='big') + \
            self.start_time.to_bytes(2, byteorder='big') + \
            self.elapse_time.to_bytes(1, byteorder='big') + \
            self.max_walking_min.to_bytes(1, byteorder='big') + \
            self.grid_size_min.to_bytes(1, byteorder='big') + \
            self.x_num.to_bytes(2, byteorder='big') + \
            self.y_num.to_bytes(2, byteorder='big') + \
            self.length.to_bytes(2, byteorder='big')



from .busSim import BusSim
from .config import Config

import logging
import sys

LOG_FILE_PATH = "busSim.log"


def _init_logger():
    logger = logging.getLogger('app')
    logger.setLevel(logging.DEBUG)

    # init file handler
    handler = logging.FileHandler(LOG_FILE_PATH)
    handler.setLevel(logging.WARNING)
    formatter = logging.Formatter(
        '%(asctime)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)

    logger.addHandler(handler)


_init_logger()



from .busSim.manager import managerFactory
from .result.searchResult import SearchResult
from .util import gen_start_time, transform
from .service.yelp import get_results
from .census import Census
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Polygon
from shapely.wkt import loads
from pyproj import Transformer
from zipfile import ZipFile
from io import TextIOWrapper
import os
from pathlib import Path
from math import ceil, floor
from collections import defaultdict
import time


class SCanalyzer:
    def __init__(self, gtfs_path):
        self.gtfs_path = gtfs_path
        self.base_out_path = self._get_out_path()
        self.out_path = self.base_out_path
        self._preprocess_gtfs()

    def set_batch_label(self, label):
        self.out_path = os.path.join(self.base_out_path, label)
        Path(self.out_path).mkdir(parents=True, exist_ok=True)

    def reset_batch_label(self):
        self.out_path = self.base_out_path

    def search(self, config, perf_df=None):
        # prerun check
        if not config.is_runnable():
            raise Exception("The current config is not runnable")

        # dynamically init a manager
        manager = managerFactory.create(
            config.get_run_env(), gtfs_path=self.gtfs_path, out_path=self.out_path, borders=self.borders)

        start_times = gen_start_time(
            config.get_interval(), config.get_busSim_params().get("elapse_time"))
        result_df = manager.run_batch(config.get_busSim_params(), start_times,
                                      config.get_start_points(), perf_df)
        return result_df

    def load_census(self, cache=True):
        """
        Looks for a stops.csv file in data/mmt_gtfs, queries TigerWeb Census API to pull out census tracts 
        based on the center and radius of the system. An optional addition of 1km (default) is added to the radius.
        From the tracts, and a default set of demographs the ACS 5-year 2019 dataset is queried to get the demographics
        data for each tract. A few statistics are computed. It returns a geodataframe with all of this information and
        saves it to the output folder. 

        cache     default=True, if true will load a saved result and return
        """

        # Pull from Cache and return:
        cache_path = os.path.join(self.base_out_path, "census.csv")
        if cache and os.path.exists(cache_path):
            census_df = pd.read_csv(cache_path)
            return self._csvdf_to_gdf(census_df)

        # Create the Geodataframe:
        c = Census(gtfs_filename="../data/mmt_gtfs/stops.csv")
        gdf_tracts = c.getCensusTracts()
        demographic_data = c.getDemographicsData(
            gdf_tracts, demographics=['Race', 'Vehicles'])

        # Save output:
        demographic_data.to_csv(cache_path, index=False)

        return self._csvdf_to_gdf(demographic_data)

    def load_yelp(self, api_key, services=["banks", "clinics", "dentists", "hospitals", "supermarket"], cache=True):
        cache_path = os.path.join(self.base_out_path, "services.csv")
        if cache and os.path.exists(cache_path):
            return pd.read_csv(cache_path)

        dfs = [get_results(api_key, service, self.borders)
               for service in services]
        df = pd.concat(dfs)
        df.to_csv(cache_path, index=False)
        return df

    def add_service_metrics(self, result_gdf, services_gdf, perf_df=None):
        # load grid size from a map_identifier (pick the first one on result_gdf)
        max_x, min_x, max_y, min_y, grid_size, x_num, y_num = self._load_grid_size(
            result_gdf)
        record_perf = (perf_df is not None)

        def get_grid(df):
            grid = np.zeros(x_num*y_num).reshape(y_num, -1)
            for index, row in df.iterrows():
                # convert to 3174 first
                x, y = transform(row["latitude"], row["longitude"])
                x_idx = floor((x - min_x) / grid_size)
                y_idx = floor((y - min_y) / grid_size)
                if x_idx >= 0 and x_idx < x_num and y_idx >= 0 and y_idx < y_num:
                    grid[y_idx][x_idx] += 1

            return [grid]

        services_grid_series = services_gdf.groupby("service").apply(get_grid)
        services_counts = defaultdict(list)
        service_perfs = []

        # loop through all map_id in result_gdf
        # for records with the same filename: group open them and pull out each bitmaps
        curr_filename = None
        grid_iter = None
        for _, row in result_gdf.iterrows():
            s = time.time()
            filename, _ = self._parse_map_identifier(row["map_identifier"])

            # check if a new file need to be open
            if filename != curr_filename:
                curr_filename = filename
                grid_iter = SearchResult.grid_iter(filename)

            grid, _ = next(grid_iter, None)
            # combine bitmaps
            for service, servicemap in services_grid_series.items():
                services_counts[service].append(0)

            for y, grid_row in enumerate(grid):
                for x, bit in enumerate(grid_row):
                    if bit == 0:
                        continue

                    for service, servicemap in services_grid_series.items():
                        services_counts[service][-1] += servicemap[0][y][x]

            service_perfs.append(time.time() - s)

        for service, col in services_counts.items():
            result_gdf[service] = col

        if record_perf:
            perf_df["add_service_time"] = service_perfs

        return result_gdf

    def add_demographic_metrics(self, result_gdf, census_gdf, perf_df=None):
        max_x, min_x, max_y, min_y, grid_size, x_num, y_num = self._load_grid_size(
            result_gdf)
        record_perf = (perf_df is not None)

        # iterate through all the starting locations (only the unique starting locations)
        start_to_demographic_dict = {}
        for result_i, row in result_gdf.iterrows():
            s = time.time()
            _, i = self._parse_map_identifier(row["map_identifier"])
            if i not in start_to_demographic_dict:
                start_to_demographic_dict[i] = np.nan
                for census_i, census_row in census_gdf.iterrows():
                    if census_row["geometry"].contains(row["geometry"]):
                        start_to_demographic_dict[i] = census_i
                        break
            pop = np.nan
            cars = np.nan
            if not np.isnan(start_to_demographic_dict[i]):
                pop = census_gdf.at[start_to_demographic_dict[i], "Tot Pop"]
                cars = census_gdf.at[start_to_demographic_dict[i],
                                     "cars per capita"]
            result_gdf.at[result_i, "Tot Pop"] = pop
            result_gdf.at[result_i, "cars per capita"] = cars

            if record_perf:
                perf_df.at[result_i, "add_census_time"] = time.time() - s

        return result_gdf

    def load_result_map(self, map_identifier):
        filename, idx = self._parse_map_identifier(map_identifier)
        grid, grid_size = SearchResult.load_grid(filename, idx)
        max_x, min_x, max_y, min_y = self.borders

        # generate gdf
        df = pd.DataFrame(
            columns=["geometry"])
        i = 0
        for y, row in enumerate(grid):
            for x, bit in enumerate(row):
                x0 = grid_size * x + min_x
                x1 = x0 + grid_size
                y0 = grid_size * y + min_y
                y1 = y0 + grid_size
                if bit == 1:
                    df.loc[i, "geometry"] = Polygon(
                        [(x0, y0), (x0, y1), (x1, y1), (x1, y0)])
                    i += 1
        gdf = gpd.GeoDataFrame(df, crs="EPSG:3174")
        return gdf

    def _get_out_path(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        out_path = os.path.join(dir_path, os.pardir, "out")
        Path(out_path).mkdir(parents=True, exist_ok=True)
        return out_path

    def _preprocess_gtfs(self):
        self._reproject_stops()
        self.borders = self._get_borders()

    def _reproject_stops(self):
        with ZipFile(self.gtfs_path) as zf:
            if "stops-3174.txt" in zf.namelist():
                return
            with zf.open("stops.txt") as f:
                stops_df = pd.read_csv(TextIOWrapper(f), sep=",")
                transformer = Transformer.from_crs(4326, 3174)
                stop_x, stop_y = transformer.transform(
                    stops_df["stop_lat"], stops_df["stop_lon"])
                stops_df["stop_x"] = stop_x
                stops_df["stop_y"] = stop_y
                # TODO change this to a fake file wrapper
                stops_df.to_csv("stops-3174.txt")

        with ZipFile(self.gtfs_path, "a") as zf:
            zf.write('stops-3174.txt')
        os.remove('stops-3174.txt')

    def _get_borders(self):
        # TODO: optimize
        # 1. combine with previous _reproject_stops to only open the file once
        # 2. these can be computed within one loop
        with ZipFile(self.gtfs_path) as zf:
            with zf.open("stops-3174.txt") as f:
                stops_df = pd.read_csv(TextIOWrapper(f), sep=",")
                max_x = stops_df["stop_x"].max()
                min_x = stops_df["stop_x"].min()
                max_y = stops_df["stop_y"].max()
                min_y = stops_df["stop_y"].min()

                return (max_x, min_x, max_y, min_y)

    def _parse_map_identifier(self, map_identifier):
        tokens = map_identifier.split("!")
        if len(tokens) != 2 or not tokens[1].isnumeric():
            raise Exception("invalid map_identifier")
        return os.path.join(self.out_path, tokens[0]), int(tokens[1])

    def _load_grid_size(self, result_gdf):
        map_identifier = result_gdf.at[0, "map_identifier"]
        filename, idx = self._parse_map_identifier(map_identifier)
        _, grid_size = SearchResult.load_grid(filename, idx)
        max_x, min_x, max_y, min_y = self.borders
        x_num = ceil(abs(max_x - min_x) / grid_size)
        y_num = ceil(abs(max_y - min_y) / grid_size)

        return max_x, min_x, max_y, min_y, grid_size, x_num, y_num

    def _csvdf_to_gdf(self, df):
        df['geometry'] = df['geometry'].apply(loads)
        gdf = gpd.GeoDataFrame(
            df, geometry="geometry", crs="EPSG:4326")
        return gdf



class ManagerFactory:
    def __init__(self):
        self._managers = {}

    def register_manager(self, key, manager):
        self._managers[key] = manager

    def get_available_managers(self):
        return list(self._managers.keys())

    def create(self, key, **kwargs):
        manager = self._managers.get(key)
        if not manager:
            raise ValueError(key)
        return manager(**kwargs)



from .baseManager import BaseManager
from ..busSim import BusSim
from ...result.searchResult import SearchResult
import os
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
from zipfile import ZipFile
import time


class LocalManager(BaseManager):
    def __init__(self, gtfs_path, out_path, borders):
        super().__init__(gtfs_path, borders)
        self.out_path = out_path

    def run_batch(self, busSim_params, start_times, start_points, perf_df=None):
        result_df = pd.DataFrame(
            columns=["geometry", "start_time", "map_identifier"])
        record_perf = (perf_df is not None)
        start_points_len = len(start_points)

        idx = 0
        for start_time in start_times:
            stop_idx = 0

            s = time.time()
            busSim = BusSim(self, busSim_params["day"], start_time, busSim_params["elapse_time"],
                            busSim_params["avg_walking_speed"], busSim_params["max_walking_min"])
            result = SearchResult(busSim, busSim_params["grid_size_min"])
            filename = result.get_out_filename()
            amortized_init_time = (time.time() - s) / start_points_len

            # run busSim search on every start_point
            for start_point in start_points:
                s = time.time()
                grid = busSim.get_access_grid(
                    start_point=start_point, grid_size_min=busSim_params["grid_size_min"], route_remove=busSim_params["route_remove"])
                result.record(start_point, grid)

                point = Point(start_point[1], start_point[0])
                # result_df.loc[idx, "area_sum"] = grid.sum()
                result_df.loc[idx, "geometry"] = point
                result_df.loc[idx, "start_time"] = start_time
                result_df.loc[idx,
                              "map_identifier"] = f"{filename}!{stop_idx}"

                if record_perf:
                    perf_df.loc[idx, "search_time"] = time.time() - \
                        s + amortized_init_time
                    perf_df.loc[idx, "geometry"] = point
                    perf_df.loc[idx, "start_time"] = start_time

                idx += 1
                stop_idx += 1

            self.save(result)

        return result_df

    def read_gtfs(self, filename):
        with ZipFile(self.gtfs_path) as zf:
            with zf.open(filename) as f:
                df = pd.read_csv(f, sep=",")
                return df

    def save(self, result):
        out_path = os.path.join(self.out_path, result.get_out_filename())
        with open(out_path, "wb") as f:
            f.write(result.to_bytes())



import requests
import pandas as pd
import geopandas as gpd
from pyproj import Transformer


def get_search_point(borders):
    max_x, min_x, max_y, min_y = borders
    # TODO: this might miss some locations, use diagonal
    radius = int(max(max_x - min_x, max_y - min_y) / 2)
    centroid_x = (max_x + min_x) / 2
    centroid_y = (max_y + min_y) / 2
    transformer = Transformer.from_crs(3174, 4326)
    centroid_lat, centroid_lon = transformer.transform(centroid_x, centroid_y)
    return radius, centroid_lat, centroid_lon


def get_results(api_key, service, borders):
    limit = 50
    endpoint = 'https://api.yelp.com/v3/businesses/search'
    headers = {'Authorization': 'bearer %s' % api_key}

    radius, centroid_lat, centroid_lon = get_search_point(borders)
    # print(radius, centroid_lat, centroid_lon)
    parameters = {
        'term': service,
        'limit': limit,
        'radius': radius,
        'latitude': centroid_lat,
        'longitude': centroid_lon}
    all_results = []
    response = requests.get(
        url=endpoint, params=parameters, headers=headers)
    result = response.json()
    # print(result)
    all_results.extend(result['businesses'])

    found = result['total']
    searches = found // limit

    if searches != 0:
        for i in range(searches):
            offset = (i+1) * limit
            parameters['offset'] = offset
            response = requests.get(
                url=endpoint, params=parameters, headers=headers)
            result = response.json()
            # print(len(all_results))
            all_results.extend(result['businesses'])
    df = pd.DataFrame(
        columns=['id', 'service', 'name', 'latitude', 'longitude'])

    for i, row in enumerate(all_results):
        # print(i)
        df.loc[i, 'id'] = row['id']
        df.loc[i, 'service'] = service
        df.loc[i, 'name'] = row['name']
        df.loc[i, 'latitude'] = row['coordinates']['latitude']
        df.loc[i, 'longitude'] = row['coordinates']['longitude']

    return df



class BaseManager():
    def __init__(self, gtfs_path, borders):
        self.gtfs_path = gtfs_path
        self.borders = borders

    def get_borders(self):
        return self.borders

    def run_batch(self, busSim_params, start_time, start_points, route_remove):
        pass

    def read_gtfs(self, filename):
        pass

    def save(self, result):
        pass



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    install_requires = fh.read()
    install_requires = install_requires.split("\n")
    install_requires = list(filter(lambda x: x[0] != "#", install_requires))

setuptools.setup(
    name="SCanalyzer",
    version="0.0.2",
    author="Chang Xu",
    author_email="cxu249@wisc.edu",
    description="Service Cube Analyzer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wisc-bus/simulator",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: GIS",
        "Programming Language :: Python :: 3",
    ],
    install_requires=install_requires,
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
)



from .baseManager import BaseManager
import os
import json
import time
from zipfile import ZipFile
import pandas as pd
from subprocess import check_output
import boto3


class AWSManager(BaseManager):
    def __init__(self, gtfs_path, city_path, out_path):
        self._s3 = boto3.client('s3')
        self._iam = boto3.client("iam")
        self._lambda = boto3.client('lambda', region_name='ap-northeast-1')
        self.bucket_name = self._create_bucket()
        # self._upload_data(gtfs_path, city_path)
        self._upload_lambda()

    def read_gtfs(self, filename):
        pass

    def read_city(self):
        pass

    def save(self, result):
        pass

    def clean_up(self):
        print("cleaning up aws resources")
        self._clean_bucket()
        self._clean_lambda()

    def _create_bucket(self):
        print("creating s3 bucket")
        account_id = self._get_account_id()
        bucket_name = f"gtfs-bussim-{account_id}"

        self._s3.create_bucket(
            Bucket=bucket_name,
        )

        return bucket_name

    def _upload_data(self, gtfs_path, city_path):
        print("uploading data")
        with open(gtfs_path, 'rb') as f:
            self._s3.put_object(Bucket=self.bucket_name,
                                Key="gtfs.zip",
                                Body=f)

    def _upload_lambda(self):
        # self._upload_lambda_layer()
        self._upload_lambda_function()

    def _upload_lambda_layer(self):
        print("packaging lambda layer")
        self.layerName = "busSim-layer"

        lambda_path = self._get_lambda_path()
        check_output(["./deploy_layer.sh"], cwd=lambda_path)
        zip_path = os.path.join(
            lambda_path, "lambda_layers", "busSim-layer.zip")

        print("deploying lambda layer")
        with open(zip_path, 'rb') as f:
            self._lambda.publish_layer_version(
                LayerName=self.layerName,
                Description='The layer needed for busSim',
                Content={
                    'ZipFile': f.read()
                },
                CompatibleRuntimes=[
                    'python3.8',
                ]
            )

    def _upload_lambda_function(self):
        self.roleName = 's3rwRole'
        self.policyName = 's3rwPolicy'
        self.functionName = 'busSim'

        # create IAM role
        print("creating IAM role")
        response = self._iam.create_role(
            RoleName=self.roleName,
            AssumeRolePolicyDocument=json.dumps({
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {
                            "Service": "lambda.amazonaws.com"
                        },
                        "Action": "sts:AssumeRole"
                    }
                ]
            })
        )
        self.roleArn = response.get("Role").get("Arn")

        # create policy
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "ListObjectsInBucket",
                    "Effect": "Allow",
                    "Action": ["s3:ListBucket"],
                    "Resource": [f"arn:aws:s3:::{self.bucket_name}"]
                },
                {
                    "Sid": "AllObjectActions",
                    "Effect": "Allow",
                    "Action": "s3:*Object",
                    "Resource": [f"arn:aws:s3:::{self.bucket_name}/*"]
                }
            ]
        }
        response = self._iam.create_policy(
            PolicyName=self.policyName,
            PolicyDocument=json.dumps(policy)
        )
        self.policyArn = response.get("Policy").get("Arn")

        # attach policy
        self._iam.attach_role_policy(
            PolicyArn=self.policyArn,
            RoleName=self.roleName
        )

        # package function
        print("packaging lambda function")
        lambda_path = self._get_lambda_path()

        tmp = 'tmp.zip'
        with ZipFile(tmp, 'w') as z:
            for name in (n for n in os.listdir(lambda_path) if n.split('.')[-1] == 'py'):
                z.write(os.path.join(lambda_path, name), '/'+name)

        print("deploying lambda function")
        time.sleep(10)
        with open(tmp, 'rb') as f:
            response = self._lambda.create_function(
                Code={
                    'ZipFile': f.read()
                },
                Description='BusSim handler',
                FunctionName=self.functionName,
                Handler='lambda_function.lambda_handler',
                MemorySize=512,
                Publish=True,
                Role=self.roleArn,
                Runtime='python3.8',
                Timeout=900,
                TracingConfig={
                    'Mode': 'Active',
                },
            )

        os.remove(tmp)

    def _clean_bucket(self):
        print("cleaning up s3")
        self._s3.delete_bucket(
            Bucket=self.bucket_name
        )

    def _clean_lambda(self):
        print("cleaning up lambda")
        self._iam.detach_role_policy(
            RoleName=self.roleName,
            PolicyArn=self.policyArn
        )

        self._iam.delete_policy(
            PolicyArn=self.policyArn
        )

        self._iam.delete_role(
            RoleName=self.roleName
        )

        self._lambda.delete_function(
            FunctionName=self.functionName
        )

    def _get_lambda_path(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        lambda_path = os.path.join(
            dir_path, os.pardir, os.pardir, os.pardir, "lambda")
        return lambda_path

    def _get_account_id(self):
        sts = boto3.client('sts')
        return sts.get_caller_identity()['Account']



import pandas as pd
from shapely.geometry import Point
from pyproj import Transformer
import geopandas as gpd
import requests
import json


class Census:
    """
    TODO
    """

    def __init__(self, gtfs_filename):
        """
        TODO
        """
        self.gtfs_file = gtfs_filename

        return

    def getDemographicsMap(self):
        """
        TODO
        """

        return self.getCensusTracts()

    def getFarthestPoints(self, filename):
        """
        TODO
        """
        lat_lon_df = pd.read_csv(filename, usecols=['stop_lat', 'stop_lon'])

        # Convert latitude and longitude points to [Shaply] Point objects
        ID_points = [Point(
            lat_lon_df['stop_lon'][i], lat_lon_df['stop_lat'][i]) for i in range(len(lat_lon_df))]

        # Find the farthest (N, E, S, W) points of the system
        farthest_points = [ID_points[0] for i in range(4)]
        for point in ID_points:
            if point.y > farthest_points[0].y:  # north = largest y (lat)
                farthest_points[0] = point
            if point.x > farthest_points[1].x:  # east = largest x (lon)
                farthest_points[1] = point
            if point.y < farthest_points[2].y:  # south = smallest y (lat)
                farthest_points[2] = point
            if point.x < farthest_points[3].x:  # west = smallest x (lon)
                farthest_points[3] = point

        return farthest_points

    def coordinateTransform(self, coordinate, old_epsg="epsg:4326", new_epsg="epsg:3857"):
        """
        TODO

        coordinate        list: shaply Point object
        old_epsg          string: old coordinate system, default = "epsg:4326", lat/lon
        new_epsg          string: new coordinate system, default = "epsg:3857", mercader special, meters
        returns           list: the transformed coordinates
        """

        transformer = Transformer.from_crs("epsg:4326", "epsg:3857")

        return transformer.transform(coordinate.y, coordinate.x)

    def getCensusTracts(self, additionalRadius=1000, old_epsg="epsg:4326", new_epsg="epsg:3857"):
        """"
        TODO
        """

        farthest_points = self.getFarthestPoints(self.gtfs_file)

        farthest_points_mercader = [self.coordinateTransform(
            point, old_epsg, new_epsg) for point in farthest_points]

        center_lon = (farthest_points_mercader[3][0] +
                      abs(farthest_points_mercader[1][0] - farthest_points_mercader[3][0])/2)
        center_lat = (farthest_points_mercader[2][1] +
                      abs(farthest_points_mercader[0][1] - farthest_points_mercader[2][1])/2)
        max_sep = max(abs(farthest_points_mercader[1][0] - farthest_points_mercader[3][0])/2,
                      abs(farthest_points_mercader[0][1] - farthest_points_mercader[2][1])/2)

        center_lon = round(center_lon, 2)
        center_lat = round(center_lat, 2)
        max_sep = round(max_sep, 2)
        max_sep + additionalRadius

        base_url = "https://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/Tracts_Blocks/MapServer/11/query?"
        geometry = "geometry=" + \
            str(round(center_lon, 2)) + "%2C" + str(round(center_lat, 2))
        mid_url = "&geometryType=esriGeometryPoint" + \
            "&spatialRel=esriSpatialRelIntersects"
        distance = "&distance=" + \
            str(round(max_sep, 2) + 2000) + "&units=esriSRUnit_Meter"
        end_url = ("&outFields=STATE%2CCOUNTY%2CTRACT%2CBLKGRP" + "&returnGeometry=true" + "&returnTrueCurves=false" +
                   "&returnIdsOnly=false" + "&returnCountOnly=false" + "&returnZ=false" + "&returnM=false" +
                   "&returnDistinctValues=false" + "&returnExtentOnly=false" + "&featureEncoding=esriDefault" +
                   "&f=geojson")

        url = base_url + geometry + mid_url + distance + end_url

        print("Getting geodataframe for tracts from Tigerweb...")
        geodataframe = gpd.read_file(url)

        return geodataframe

    def getDemographicsData(self, CensusTracts, demographics=['Race'], demographicsDefaults=True, sample=False):
        """"
        TODO
        if demographicsDefaults==True: then demographics should be a list of strings for demographis to return
        else: demographics should be a list of dictionaries specifying the codes from the census
        """

        Race = {'Total': 'B02001_001E', 'White alone': 'B02001_002E', 'Black or African American alone': 'B02001_003E',
                'American Indian and Alaska Native alone': 'B02001_004E', 'Asian alone': 'B02001_005E',
                'Native Hawaiian and Other Pacific Islander alone': 'B02001_006E'}

        Vehicles = {'Total Vehicles': 'B25044_001E'}

        print(
            "Getting demographics data from Census.gov, this may take a couple minutes...")

        combined_data = []

        if sample:
            CensusTracts = CensusTracts[0:10]

        for row in CensusTracts.iterrows():
            state = row[1]['STATE']
            tract = row[1]['TRACT']
            county = row[1]['COUNTY']
            blockgroup = row[1]['BLKGRP']

            race_list = [Race[key] for key in Race]
            car_list = [Vehicles[key] for key in Vehicles]
            demographics = race_list + car_list
            demographics = ','.join(demographics)

            url_acs = "https://api.census.gov/data/2019/acs/acs5?"
            url_demographic = "get=NAME," + str(demographics) + "&"
            url_location = ("for=block%20group:" + str(blockgroup) + "&in=state:" + str(state) +
                            "%20county:" + str(county) + "%20tract:" + str(tract))
            url = url_acs + url_demographic + url_location

            resp = requests.get(url)
            data = json.loads(resp.text)

            if len(combined_data) == 0:
                combined_data.append(data[0])
            combined_data.append(data[1])

        combined_df = pd.DataFrame(combined_data[1:], columns=combined_data[0])
        combined_df.rename(columns={'tract': 'TRACT', 'state': 'STATE',
                                    'county': 'COUNTY', 'block group': 'BLKGRP'}, inplace=True)

        joined_df = CensusTracts.merge(
            combined_df, on=['TRACT', 'STATE', 'COUNTY', 'BLKGRP'])

        print(type(joined_df))
        # Id the demographics are the defaults then the stats will be computed and added to the geodataframe,
        # if it is not default then the user will have to compute them after the gdf is returned but before it
        # is combined with the results from the simulator
        if demographicsDefaults:

            joined_df['Tot Pop'] = joined_df['B02001_001E'].astype(int)
            joined_df['% White alone'] = (joined_df['B02001_002E'].astype(
                int) / joined_df['B02001_001E'].astype(int))
            joined_df['% Black or African American alone'] = (joined_df['B02001_003E'].astype(int) /
                                                              joined_df['B02001_001E'].astype(int))
            joined_df['% American Indian and Alaska Native alone'] = (joined_df['B02001_004E'].astype(int) /
                                                                      joined_df['B02001_001E'].astype(int))
            joined_df['% Asian alone alone'] = (joined_df['B02001_005E'].astype(int) /
                                                joined_df['B02001_001E'].astype(int))
            joined_df['% Native Hawaiian and Other Pacific Islander alone'] = (joined_df['B02001_006E'].astype(int) /
                                                                               joined_df['B02001_001E'].astype(int))

            joined_df['cars per capita'] = (joined_df['B25044_001E'].astype(
                int) / joined_df['B02001_001E'].astype(int))

        return joined_df



import pandas as pd
from datetime import datetime
from pyproj import CRS, Transformer
import os
import random

# for quick debugging
debug = True


def dprint(cmd):
    if debug:
        print(cmd)


def fmin(minutes):
    return f"{minutes // 60}:{minutes % 60}:00"


def tomin(time_string):
    time_li = time_string.split(":")
    h = int(time_li[0])
    m = int(time_li[1])
    return h * 60 + m


def gen_start_time(interval, elapse_time):
    start_times = []
    curr = 0
    interval = tomin(interval)
    elapse_time = tomin(elapse_time)
    min_total = 24 * 60 - elapse_time
    while curr <= min_total:
        start_times.append(fmin(curr))
        curr += interval

    return start_times


def gen_locations(data_path, num):
    mmt_gtfs_path = os.path.join(data_path, "mmt_gtfs")
    stops_df = pd.read_csv(os.path.join(
        mmt_gtfs_path, "stops.csv"), sep=",")
    locations = []

    lat_min = stops_df.stop_lat.min()
    lat_max = stops_df.stop_lat.max()
    lon_min = stops_df.stop_lon.min()
    lon_max = stops_df.stop_lon.max()

    for i in range(num):
        lat = random.random() * (lat_max - lat_min) + lat_min
        lon = random.random() * (lon_max - lon_min) + lon_min
        locations.append((lat, lon))

    return locations


# for parsing perf logs
def parse_log_line(line):
    line = line.strip()
    tokens = line.split(" - ")
    timestamp = datetime.strptime(tokens[0], "%Y-%m-%d %H:%M:%S,%f")
    level = tokens[1]
    msg = tokens[2]
    return timestamp, level, msg


def to_millisecs(timedelta):
    return timedelta.total_seconds() * 1000


def get_perf(log_path):

    perf = pd.DataFrame()
    graph_search = projections = uninioning = None
    graph_search_ts = projections_ts = uninioning_ts = None
    i = 0

    with open(log_path) as f:
        for line in f:
            timestamp, level, msg = parse_log_line(line)

            if msg == "Start searching graph":
                # print(1)
                graph_search_ts = timestamp
            elif msg == "start generating gdf":
                # print(2)
                projections_ts = timestamp
                graph_search = projections_ts - graph_search_ts
            elif msg == "start calculating union/difference":
                # print(3)
                uninioning_ts = timestamp
                projections = uninioning_ts - projections_ts
            elif msg == "finish calculating area":
                # print(4)
                uninioning = timestamp - uninioning_ts
                # flush all previous results
                perf.loc[i, "graph_search"] = to_millisecs(graph_search)
                perf.loc[i, "projections"] = to_millisecs(projections)
                perf.loc[i, "uninioning"] = to_millisecs(uninioning)
                graph_search = projections = uninioning = None
                graph_search_ts = projections_ts = uninioning_ts = None
                i += 1

    return perf


# for transforming CRS
# https://epsg.io/3174
_transformer = None


def transform(lat, lon):
    global _transformer
    if _transformer is None:
        _transformer = Transformer.from_crs(4326, 3174)
    return _transformer.transform(lat, lon)

# for serializing grid to bitmap string


def serialize_grid():
    pass


def deserialize_grid():
    pass



import zipfile, shutil, csv, io, itertools, sys
from collections import namedtuple

Trip = namedtuple("Trip", ["row", "stop_times"])

def edit_half(trips_hdr, stop_times_hdr, trips, trip_id_gen):
    trips_new = trips[::2]
    trips.clear()
    trips.extend(trips_new)

def copy_with_edits(path1, path2, edit_fn, route):
    with zipfile.ZipFile(path1) as zf1, zipfile.ZipFile(path2, "w") as zf2:
        # STEP 1: copy the files that don't get changed
        for info in zf1.infolist():
            if info.filename in ("trips.txt", "stop_times.txt"):
                continue
            with zf1.open(info.filename) as f1, zf2.open(info.filename, "w") as f2:
                shutil.copyfileobj(f1, f2)

        # STEP 2: separate trips that might get changed, copy trips that will not
        trips_changed = {} # key=trip_id, val=Trip object
        trips_unchanged = []
        max_trip_id = 0
        with zf1.open("trips.txt") as f:
            r = csv.reader(io.TextIOWrapper(f))
            trips_hdr = next(r)
            route_idx, trip_id_idx = trips_hdr.index("route_short_name"), trips_hdr.index("trip_id")
            for trip in r:
                trip_id = trip[trip_id_idx]
                if int(trip_id) > max_trip_id:
                    max_trip_id = int(trip_id)

                if int(trip[route_idx]) == int(route):
                    trips_changed[trip_id] = Trip(trip, [])
                else:
                    trips_unchanged.append(trip)

        with zf1.open("stop_times.txt") as f1, zf2.open("stop_times.txt", "w") as f2:
            # STEP 3: separate stop_times that might get changed, copy those that will not
            r = csv.reader(io.TextIOWrapper(f1))
            w = csv.writer(io.TextIOWrapper(f2, write_through=True))
            stimes_hdr = next(r)
            w.writerow(stimes_hdr)
            trip_id_idx = stimes_hdr.index("trip_id")

            for stime in r:
                trip_id = stime[trip_id_idx]
                if trip_id in trips_changed:
                    trips_changed[trip_id].stop_times.append(stime)
                else:
                    w.writerow(stime)

            # STEP 4: callback to determine edits
            trips_changed_lst = list(trips_changed.values())
            edit_fn(trips_hdr, stimes_hdr, trips_changed_lst,
                    itertools.count(max_trip_id+1))

            # STEP 5: copy over modified stop_times
            for trip in trips_changed_lst:
                for stime in trip.stop_times:
                    w.writerow(stime)

        # STEP 6: copy over modified trips
        with zf2.open("trips.txt", "w") as f:
            w = csv.writer(io.TextIOWrapper(f, write_through=True))
            w.writerow(trips_hdr)
            for trip in trips_unchanged:
                w.writerow(trip)
            for trip in trips_changed_lst:
                w.writerow(trip.row)

def main():
    if len(sys.argv) != 5:
        print("Usage: python3 gtfs_edit.py <orig-gtfs.zip> <new-gtfs.zip> <operation> <route>")
        print("Example: python3 gtfs_edit.py mmt_gtfs.zip out.zip half 80")
        sys.exit(1)

    path1, path2, op, route = sys.argv[1:]
    edit_fns = {"half": edit_half}
    if not op in edit_fns:
        print(f"{op} is not a valid operation.  Choose one of these:")
        print(",".join(sorted(edit_fns.keys())))
        sys.exit(1)

    copy_with_edits(path1, path2, edit_fns[op], route)

if __name__ == '__main__':
     main()
